/* -------------------------------------------------------- */
/*             == DOL SPLIT UTILITY v. 1.0 ! ==             */
/* -------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// DOLPHIN used types
typedef signed char         s8;
typedef signed short        s16;
typedef signed long         s32;
typedef signed __int64      s64;
typedef unsigned char       u8;
typedef unsigned short      u16;
typedef unsigned long       u32;
typedef unsigned __int64    u64;

// DOL header
#define DOL_NUM_TEXT    7
#define DOL_NUM_DATA    11

typedef struct
{
    u32     textOffset[DOL_NUM_TEXT];
    u32     dataOffset[DOL_NUM_DATA];

    u32     textAddress[DOL_NUM_TEXT];
    u32     dataAddress[DOL_NUM_DATA];

    u32     textSize[DOL_NUM_TEXT];
    u32     dataSize[DOL_NUM_DATA];

    u32     bssAddress;
    u32     bssSize;
    u32     entryPoint;
    u32     padd[7];
} DolHeader;

// command string options
int command, section, numsect;
char dolname[256], filename[256];

void ban()
{
    printf("GAMECUBE DOL executable split utility\n");
    printf("(c) org %s (mailto:kvzorganic@mail.ru)\n", __DATE__);
    printf("\n");
}

void help()
{
    printf("USE : dolsplit <add|rem|show|help> <all|code|data> [n] <dolname> {filename}\n");
    printf("\n");
    printf("  add  : add \"code\" or \"data\" section as \"n\" from \"filename\",\n");
    printf("         section will be replaced, if already exist.\n");
    printf("  rem  : remove existed \"n\" section from \"code\" or \"data\",\n");
    printf("         remove \"all\" sections,\n");
//  printf("         \"filename\" can be specified for single section.\n");
    printf("  show : show section \"n\" from \"code\" or \"data\",\n");
    printf("         show all sections from \"code\" or \"data\",\n");
    printf("         show them \"all\"\n");
    printf("  help : additional help\n");
    printf("\n");
}

// additional help
void dol_help()
{
    printf("DOL header fields description :\n");
    printf("   addr - virtual address to load section (memory address)\n");
    printf("   offs - offset of section, from begin of the DOL file\n");
    printf("   size - size of section, in bytes, must be 32-byte aligned\n");
    printf("\n");
    printf("examples of usage : \n");
    printf("   dolsplit show all default.dol         - show overall DOL\n");
    printf("   dolsplit show all 2 default.dol       - show only #2 sections\n");
    printf("   dolsplit rem data default.dol         - remove all data sections\n");
    printf("   dolsplit rem code 4 default.dol       - remove specified code section\n");
    printf("   dolsplit rem all default.dol          - clean-up DOL\n");
    printf("   dolsplit add code 0 default.dol a.bin - add code section #0\n");
    printf("   dolsplit add data 2 default.dol b.dat - add data section #2\n");
    printf("\n");
    printf("note : code sections are limited to 0...6, data limited to 0...10.\n");
    printf("\n");
}

void revendian(void *ptr, long count)
{
    unsigned long *ulptr = (unsigned long *)ptr;

    count >>= 2;

    while(count--)
    {
        unsigned long data = *ulptr;

        unsigned char 
            b1 = (unsigned char)(data      ) & 0xff,
            b2 = (unsigned char)(data >>  8) & 0xff,
            b3 = (unsigned char)(data >> 16) & 0xff,
            b4 = (unsigned char)(data >> 24) & 0xff;

        *ulptr = 
            ((unsigned long)b1 << 24) |
            ((unsigned long)b2 << 16) |
            ((unsigned long)b3 <<  8) | b4;

        ulptr++;
    }
}

// remove block from the file (WIN32 only)
void frem(char *fname, long offs, long size)
{
    FILE *f;
    unsigned char *oldbuf = NULL;
    long fsize;

    f = fopen(fname, "rb");
    if(!f) return;

    fseek(f, 0, SEEK_END);
    fsize = ftell(f);
    fseek(f, 0, SEEK_SET);

    if(offs >= fsize)
    {
        fclose(f);
        return;
    }

    if((offs + size) > fsize) size = fsize - offs;

    oldbuf = malloc(fsize);
    if(!oldbuf)
    {
        fclose(f);
        return;
    }
    else
    {
        fread(oldbuf, 1, fsize, f);
        fclose(f);
    }

    f = fopen(fname, "wb");
    if(!f)
    {
        free(oldbuf);
        return;
    }

    fwrite(oldbuf, 1, offs, f);
    fwrite(&oldbuf[offs + size], 1, fsize - (offs + size), f);

    fclose(f);
    free(oldbuf);
}

// concat two files : file1 = file1 + file2
// return length of file2, if success
long fcat(char *file1, char *file2)
{
    FILE *f1, *f2;
    unsigned char *buf = NULL;
    long size1, size2;

    f1 = fopen(file1, "rb");
    if(!f1) return 0;

    fseek(f1, 0, SEEK_END);
    size1 = ftell(f1);
    fseek(f1, 0, SEEK_SET);

    f2 = fopen(file2, "rb");
    if(!f2) return 0;

    fseek(f2, 0, SEEK_END);
    size2 = ftell(f2);
    fseek(f2, 0, SEEK_SET);

    buf = malloc(size2);
    if(!buf)
    {
        fclose(f1);
        fclose(f2);
        return 0;
    }

    fread(buf, 1, size2, f2);
    fclose(f2);
    fclose(f1);
    f1 = fopen(file1, "ab");
    fwrite(buf, 1, size2, f1);
    free(buf);

    return size2;
}

// write back modified DOL header (one-shot function!)
void dol_update(DolHeader *dol)
{
    FILE *f = fopen(dolname, "rb");
    long size;
    unsigned char *buf;
    
    revendian(dol, sizeof(DolHeader));

    fseek(f, 0, SEEK_END);
    size = ftell(f);
    fseek(f, 0, SEEK_SET);

    buf = malloc(size);
    fread(buf, 1, size, f);
    memcpy(buf, dol, sizeof(DolHeader));
    fclose(f);

    f = fopen(dolname, "wb");
    fwrite(buf, 1, size, f);
    
    free(buf);
    fclose(f);
}

void main(int argc, char **argv)
{
    int i, entry;
    DolHeader dol;
    FILE *f;

    ban();

    // show help, if no arguments passed
    if(argc <= 1)
    {
        help();
        return;
    }

    // get command
    if(!strcmp(argv[1], "add")) command = 1;
    else if(!strcmp(argv[1], "rem")) command = 2;
    else if(!strcmp(argv[1], "show")) command = 3;
    else if(!strcmp(argv[1], "help")) { dol_help(); return; }
    else command = 0;

    // parse command
    switch(command)
    {
        // ------------------- ADD -------------------- //

        case 1:
            if(argc <= 2)
            {
                printf("specify sections to add : code or data.\n");
                return;
            }

            // get section
            if(!strcmp(argv[2], "all")) section = 0;
            else if(!strcmp(argv[2], "code")) section = 2;
            else if(!strcmp(argv[2], "data")) section = 3;
            else section = 0;

            if(argc <= 3)
            {
                help();
                return;
            }

            // get section number and DOL name
            if(strrchr(argv[3], '.'))
            {
                strcpy(dolname, argv[3]);
                numsect = -1;
            }
            else
            {
                numsect = atoi(argv[3]);

                if(argc <= 4)
                {
                    help();
                    return;
                }
                else strcpy(dolname, argv[4]);
            }

            // load DOL header
            f = fopen(dolname, "rb");
            if(!f)
            {
                printf("cannot open DOL : %s\n", dolname);
                return;
            }
            fread(&dol, 1, sizeof(DolHeader), f);
            revendian(&dol, sizeof(DolHeader));
            fclose(f);

            // get filename
            if(numsect == -1) { if(argc >= 5) strcpy(filename, argv[4]); else filename[0] = 0; }
            else { if(argc >= 6) strcpy(filename, argv[5]); else filename[0] = 0; }

            switch(section)
            {
                case 2:         // code
                    if((numsect < 0) || (numsect >= DOL_NUM_TEXT))
                    {
                        printf("specify correct code section number (in range 0...%i).\n", DOL_NUM_TEXT - 1);
                    }
                    {
                        // check filename
                        if(strlen(filename) == 0)
                        {
                            printf("specify \"filename\" for section.\n");
                            break;
                        }
                        else
                        {
                            FILE *bin = fopen(filename, "rb");
                            if(!bin)
                            {
                                printf("cannot open specified \"filename\" : %s.\n", filename);
                                break;
                            }
                            else fclose(bin);
                        }

                        // remove existed
                        if(dol.textAddress[numsect])
                        {
                            frem(dolname, dol.textOffset[numsect], dol.textSize[numsect]);
                                
                            // correct
                            for(entry=0; entry<DOL_NUM_TEXT; entry++)
                            {
                                if(dol.textOffset[entry] > dol.textOffset[numsect])
                                    dol.textOffset[entry] -= dol.textSize[numsect];
                            }
                            for(entry=0; entry<DOL_NUM_DATA; entry++)
                            {
                                if(dol.dataOffset[entry] > dol.textOffset[numsect])
                                    dol.dataOffset[entry] -= dol.textSize[numsect];
                            }

                            dol.textAddress[numsect] =
                            dol.textOffset[numsect] = 
                            dol.textSize[numsect] = 0;
                        }

                        // split
                        f = fopen(dolname, "rb");
                        fseek(f, 0, SEEK_END);
                        dol.textOffset[numsect] = ftell(f); 
                        fclose(f);
                        dol.textSize[numsect] = fcat(dolname, filename);

                        // input section address
                        {
                            char str[64], *nused;
                            u32 addr;
                            printf("input section address : ");
                            scanf("%s", str);
                            addr = strtoul(str, &nused, 16);
                            dol.textAddress[numsect] = addr;
                        }

                        dol_update(&dol);
                    }
                    break;

                case 3:         // data
                    if((numsect < 0) || (numsect >= DOL_NUM_DATA))
                    {
                        printf("specify correct data section number (in range 0...%i).\n", DOL_NUM_DATA - 1);
                    }
                    else
                    {
                        // check filename
                        if(strlen(filename) == 0)
                        {
                            printf("specify \"filename\" for section.\n");
                            break;
                        }
                        else
                        {
                            FILE *bin = fopen(filename, "rb");
                            if(!bin)
                            {
                                printf("cannot open specified \"filename\" : %s.\n", filename);
                                break;
                            }
                            else fclose(bin);
                        }

                        // remove existed
                        if(dol.dataAddress[numsect])
                        {
                            frem(dolname, dol.dataOffset[numsect], dol.dataSize[numsect]);
                                
                            // correct
                            for(entry=0; entry<DOL_NUM_TEXT; entry++)
                            {
                                if(dol.textOffset[entry] > dol.dataOffset[numsect])
                                    dol.textOffset[entry] -= dol.dataSize[numsect];
                            }
                            for(entry=0; entry<DOL_NUM_DATA; entry++)
                            {
                                if(dol.dataOffset[entry] > dol.dataOffset[numsect])
                                    dol.dataOffset[entry] -= dol.dataSize[numsect];
                            }

                            dol.dataAddress[numsect] =
                            dol.dataOffset[numsect] = 
                            dol.dataSize[numsect] = 0;
                        }

                        // split
                        f = fopen(dolname, "rb");
                        fseek(f, 0, SEEK_END);
                        dol.dataOffset[numsect] = ftell(f); 
                        fclose(f);
                        dol.dataSize[numsect] = fcat(dolname, filename);

                        // input section address
                        {
                            char str[64], *nused;
                            u32 addr;
                            printf("input section address : ");
                            scanf("%s", str);
                            addr = strtoul(str, &nused, 16);
                            dol.dataAddress[numsect] = addr;
                        }

                        dol_update(&dol);
                    }
                    break;

                default:
                    printf("unknown section type : %s. must be code or data.\n", argv[2]);
            }
            break;

        // ------------------- REM -------------------- //

        case 2:
            if(argc <= 2)
            {
                printf("specify sections to remove : all, code or data.\n");
                return;
            }

            // get section
            if(!strcmp(argv[2], "all")) section = 1;
            else if(!strcmp(argv[2], "code")) section = 2;
            else if(!strcmp(argv[2], "data")) section = 3;
            else section = 0;

            if(argc <= 3)
            {
                help();
                return;
            }

            // get section number and DOL name
            if(strrchr(argv[3], '.'))
            {
                strcpy(dolname, argv[3]);
                numsect = -1;
            }
            else
            {
                numsect = atoi(argv[3]);

                if(argc <= 4)
                {
                    help();
                    return;
                }
                else strcpy(dolname, argv[4]);
            }

            // load DOL header
            f = fopen(dolname, "rb");
            if(!f)
            {
                printf("cannot open DOL : %s\n", dolname);
                return;
            }
            fread(&dol, 1, sizeof(DolHeader), f);
            revendian(&dol, sizeof(DolHeader));
            fclose(f);

            // get filename
            if(numsect == -1) { if(argc >= 5) strcpy(filename, argv[4]); else filename[0] = 0; }
            else { if(argc >= 6) strcpy(filename, argv[5]); else filename[0] = 0; }

            switch(section)
            {
                case 1:         // all
                    if(numsect == -1)
                    {
                        numsect = DOL_NUM_DATA;

                        for(i=0; i<numsect; i++)
                        {
                            if(dol.textAddress[i] == 0) continue;
                            if(i < DOL_NUM_TEXT)
                            {
                                frem(dolname, dol.textOffset[i], dol.textSize[i]);
                                
                                // correct
                                for(entry=0; entry<DOL_NUM_TEXT; entry++)
                                {
                                    if(dol.textOffset[entry] > dol.textOffset[i])
                                        dol.textOffset[entry] -= dol.textSize[i];
                                }
                                for(entry=0; entry<DOL_NUM_DATA; entry++)
                                {
                                    if(dol.dataOffset[entry] > dol.textOffset[i])
                                        dol.dataOffset[entry] -= dol.textSize[i];
                                }

                                dol.textAddress[i] = 
                                dol.textOffset[i] = 
                                dol.textSize[i] = 0;
                            }
                        }

                        for(i=0; i<numsect; i++)
                        {
                            if(dol.dataAddress[i] == 0) continue;
                            frem(dolname, dol.dataOffset[i], dol.dataSize[i]);

                            // correct
                            for(entry=0; entry<DOL_NUM_TEXT; entry++)
                            {
                                if(dol.textOffset[entry] > dol.dataOffset[i])
                                    dol.textOffset[entry] -= dol.dataSize[i];
                            }
                            for(entry=0; entry<DOL_NUM_DATA; entry++)
                            {
                                if(dol.dataOffset[entry] > dol.dataOffset[i])
                                    dol.dataOffset[entry] -= dol.dataSize[i];
                            }

                            dol.dataAddress[i] = 
                            dol.dataOffset[i] = 
                            dol.dataSize[i] = 0;
                        }
                    }
                    else
                    {
                        for(i=0; i<DOL_NUM_TEXT; i++)
                        {
                            if((i == numsect) && (dol.textAddress[i]))
                            {
                                frem(dolname, dol.textOffset[i], dol.textSize[i]);

                                // correct
                                for(entry=0; entry<DOL_NUM_TEXT; entry++)
                                {
                                    if(dol.textOffset[entry] > dol.textOffset[i])
                                        dol.textOffset[entry] -= dol.textSize[i];
                                }
                                for(entry=0; entry<DOL_NUM_DATA; entry++)
                                {
                                    if(dol.dataOffset[entry] > dol.textOffset[i])
                                        dol.dataOffset[entry] -= dol.textSize[i];
                                }

                                dol.textAddress[i] = 
                                dol.textOffset[i] = 
                                dol.textSize[i] = 0;
                            }
                        }

                        for(i=0; i<DOL_NUM_DATA; i++)
                        {
                            if((i == numsect) && (dol.dataAddress[i]))
                            {
                                frem(dolname, dol.dataOffset[i], dol.dataSize[i]);

                                // correct
                                for(entry=0; entry<DOL_NUM_TEXT; entry++)
                                {
                                    if(dol.textOffset[entry] > dol.dataOffset[i])
                                        dol.textOffset[entry] -= dol.dataSize[i];
                                }
                                for(entry=0; entry<DOL_NUM_DATA; entry++)
                                {
                                    if(dol.dataOffset[entry] > dol.dataOffset[i])
                                        dol.dataOffset[entry] -= dol.dataSize[i];
                                }

                                dol.dataAddress[i] = 
                                dol.dataOffset[i] = 
                                dol.dataSize[i] = 0;
                            }
                        }
                    }

                    dol_update(&dol);
                    break;

                case 2:         // code
                    if(numsect == -1)
                    {
                        numsect = DOL_NUM_TEXT;

                        for(i=0; i<numsect; i++)
                        {
                            if(dol.textAddress[i] == 0) continue;
                            frem(dolname, dol.textOffset[i], dol.textSize[i]);

                            // correct
                            for(entry=0; entry<DOL_NUM_TEXT; entry++)
                            {
                                if(dol.textOffset[entry] > dol.textOffset[i])
                                    dol.textOffset[entry] -= dol.textSize[i];
                            }
                            for(entry=0; entry<DOL_NUM_DATA; entry++)
                            {
                                if(dol.dataOffset[entry] > dol.textOffset[i])
                                    dol.dataOffset[entry] -= dol.textSize[i];
                            }

                            dol.textAddress[i] = 
                            dol.textOffset[i] = 
                            dol.textSize[i] = 0;
                        }
                    }
                    else
                    {
                        for(i=0; i<DOL_NUM_TEXT; i++)
                        {
                            if((i == numsect) && (dol.textAddress[i]))
                            {
                                frem(dolname, dol.textOffset[i], dol.textSize[i]);

                                // correct
                                for(entry=0; entry<DOL_NUM_TEXT; entry++)
                                {
                                    if(dol.textOffset[entry] > dol.textOffset[i])
                                        dol.textOffset[entry] -= dol.textSize[i];
                                }
                                for(entry=0; entry<DOL_NUM_DATA; entry++)
                                {
                                    if(dol.dataOffset[entry] > dol.textOffset[i])
                                        dol.dataOffset[entry] -= dol.textSize[i];
                                }

                                dol.textAddress[i] = 
                                dol.textOffset[i] = 
                                dol.textSize[i] = 0;
                            }
                        }
                    }

                    dol_update(&dol);
                    break;

                case 3:         // data
                    if(numsect == -1)
                    {
                        numsect = DOL_NUM_DATA;

                        for(i=0; i<numsect; i++)
                        {
                            if(dol.dataAddress[i] == 0) continue;
                            frem(dolname, dol.dataOffset[i], dol.dataSize[i]);

                            // correct
                            for(entry=0; entry<DOL_NUM_TEXT; entry++)
                            {
                                if(dol.textOffset[entry] > dol.dataOffset[i])
                                    dol.textOffset[entry] -= dol.dataSize[i];
                            }
                            for(entry=0; entry<DOL_NUM_DATA; entry++)
                            {
                                if(dol.dataOffset[entry] > dol.dataOffset[i])
                                    dol.dataOffset[entry] -= dol.dataSize[i];
                            }

                            dol.dataAddress[i] = 
                            dol.dataOffset[i] = 
                            dol.dataSize[i] = 0;
                        }
                    }
                    else
                    {
                        for(i=0; i<DOL_NUM_DATA; i++)
                        {
                            if((i == numsect) && (dol.dataAddress[i]))
                            {
                                frem(dolname, dol.dataOffset[i], dol.dataSize[i]);

                                // correct
                                for(entry=0; entry<DOL_NUM_TEXT; entry++)
                                {
                                    if(dol.textOffset[entry] > dol.dataOffset[i])
                                        dol.textOffset[entry] -= dol.dataSize[i];
                                }
                                for(entry=0; entry<DOL_NUM_DATA; entry++)
                                {
                                    if(dol.dataOffset[entry] > dol.dataOffset[i])
                                        dol.dataOffset[entry] -= dol.dataSize[i];
                                }

                                dol.dataAddress[i] = 
                                dol.dataOffset[i] = 
                                dol.dataSize[i] = 0;
                            }
                        }
                    }

                    dol_update(&dol);
                    break;

                default:
                    printf("unknown section type : %s. must be all, code or data.\n", argv[2]);
            }
            break;

        // ------------------- SHOW ------------------- //

        case 3:
            if(argc <= 2)
            {
                printf("specify sections to show : all, code or data.\n");
                return;
            }

            // get section
            if(!strcmp(argv[2], "all")) section = 1;
            else if(!strcmp(argv[2], "code")) section = 2;
            else if(!strcmp(argv[2], "data")) section = 3;
            else section = 0;

            if(argc <= 3)
            {
                help();
                return;
            }

            // get section number and DOL name
            if(strrchr(argv[3], '.'))
            {
                strcpy(dolname, argv[3]);
                numsect = -1;
            }
            else
            {
                numsect = atoi(argv[3]);

                if(argc <= 4)
                {
                    help();
                    return;
                }
                else strcpy(dolname, argv[4]);
            }

            // load DOL header
            f = fopen(dolname, "rb");
            if(!f)
            {
                printf("cannot open DOL : %s\n", dolname);
                return;
            }
            fread(&dol, 1, sizeof(DolHeader), f);
            revendian(&dol, sizeof(DolHeader));
            fclose(f);

            switch(section)
            {
                case 1:         // all
                    if(numsect == -1)
                    {
                        numsect = DOL_NUM_DATA;

                        for(i=0; i<numsect; i++)
                        {
                            if(dol.textAddress[i] == 0) continue;
                            if(i < DOL_NUM_TEXT)
                            {
                                printf("code %i : addr=%.8X size=%.8X offs=%.8X\n", 
                                    i, dol.textAddress[i], dol.textSize[i], dol.textOffset[i]);
                            }
                        }

                        for(i=0; i<numsect; i++)
                        {
                            if(dol.dataAddress[i] == 0) continue;
                            printf("data %i : addr=%.8X size=%.8X offs=%.8X\n", 
                                i, dol.dataAddress[i], dol.dataSize[i], dol.dataOffset[i]);
                        }
                    }
                    else
                    {
                        for(i=0; i<DOL_NUM_TEXT; i++)
                        {
                            if((i == numsect) && (dol.textAddress[i]))
                            {
                                printf("code %i : addr=%.8X size=%.8X offs=%.8X\n", 
                                    i, dol.textAddress[i], dol.textSize[i], dol.textOffset[i]);
                            }
                        }

                        for(i=0; i<DOL_NUM_DATA; i++)
                        {
                            if((i == numsect) && (dol.dataAddress[i]))
                            {
                                printf("data %i : addr=%.8X size=%.8X offs=%.8X\n", 
                                    i, dol.dataAddress[i], dol.dataSize[i], dol.dataOffset[i]);
                            }
                        }
                    }
                    break;

                case 2:         // code
                    if(numsect == -1)
                    {
                        numsect = DOL_NUM_TEXT;

                        for(i=0; i<numsect; i++)
                        {
                            if(dol.textAddress[i] == 0) continue;
                            printf("code %i : addr=%.8X size=%.8X offs=%.8X\n", 
                                i, dol.textAddress[i], dol.textSize[i], dol.textOffset[i]);
                        }
                    }
                    else
                    {
                        for(i=0; i<DOL_NUM_TEXT; i++)
                        {
                            if((i == numsect) && (dol.textAddress[i]))
                            {
                                printf("code %i : addr=%.8X size=%.8X offs=%.8X\n", 
                                    i, dol.textAddress[i], dol.textSize[i], dol.textOffset[i]);
                            }
                        }
                    }
                    break;

                case 3:         // data
                    if(numsect == -1)
                    {
                        numsect = DOL_NUM_DATA;

                        for(i=0; i<numsect; i++)
                        {
                            if(dol.dataAddress[i] == 0) continue;
                            printf("data %i : addr=%.8X size=%.8X offs=%.8X\n", 
                                i, dol.dataAddress[i], dol.dataSize[i], dol.dataOffset[i]);
                        }
                    }
                    else
                    {
                        for(i=0; i<DOL_NUM_DATA; i++)
                        {
                            if((i == numsect) && (dol.dataAddress[i]))
                            {
                                printf("data %i : addr=%.8X size=%.8X offs=%.8X\n", 
                                    i, dol.dataAddress[i], dol.dataSize[i], dol.dataOffset[i]);
                            }
                        }
                    }
                    break;

                default:
                    printf("unknown section type : %s. must be all, code or data.\n", argv[2]);
            }
            break;

        default:
            printf("unknown command : %s. must be add, rem or show.\n", argv[1]);
    }
}
